""" Gravitational Wave Surrogate classes for text and hdf5 files"""

from __future__ import division  # for py2

__copyright__ = "Copyright (C) 2014 Scott Field and Chad Galley"
__email__     = "sfield@astro.cornell.edu, crgalley@tapir.caltech.edu"
__status__    = "testing"
__author__    = "Jonathan Blackman, Scott Field, Chad Galley, Vijay Varma, Kevin Barkett"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

# adding "_" prefix to potentially unfamiliar module names
# so they won't show up in gws' tab completion
import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline as _iuspline
from gwtools.harmonics import sYlm as _sYlm

# setting __package__ to gwsurrogate.new so relative imports work
__package__="gwsurrogate.new"

# assumes unique global names
from .saveH5Object import SimpleH5Object
from .saveH5Object import H5ObjectList
from .saveH5Object import H5ObjectDict
from .nodeFunction import NodeFunction
from .spline_evaluation import TensorSplineGrid, fast_complex_tensor_spline_eval
from gwsurrogate import spline_interp_Cwrapper
from .tidal_functions import UniversalRelationLambda2ToI, \
    UniversalRelationLambda2ToOmega2, UniversalRelationLambda2ToLambda3, \
    UniversalRelationLambda3ToOmega3, UniversalRelationLambda2ToAqm, \
    EffectiveDeformabilityFromDynamicalTides, PNT2Tidal, \
    EffectiveDissipativeDynamicalTides, StrainTidalEnhancementFactor

PARAM_NUDGE_TOL = 1.e-12 # Default relative tolerance for nudging edge cases


def _identity(r1, r2):
    return r1, r2
def _amp_phase(r1, r2):
    return r1['amp']*np.exp(1.j*r1['phase'])
def _re_im(r1, r2):
    return r1['re'] + 1.j*r1['im']

RECOMBINATION_FUNCS = {
    'identity': _identity,
    'amp_phase': _amp_phase,
    're_im': _re_im,
        }


def _mode_sum(modes, theta, phi):
    h = 0.
    for (ell, m), h_mode in modes.items(): # inefficient in py2
        h += _sYlm(-2, ell, m, theta, phi) * h_mode
    return h


def _splinterp(xout, xin, yin, k=3, ext='const'):
    """Uses InterpolatedUnivariateSpline to interpolate real or complex data"""
    if np.iscomplexobj(yin):
        re = _splinterp(xout, xin, np.real(yin), k=k, ext=ext)
        im = _splinterp(xout, xin, np.imag(yin), k=k, ext=ext)
        return re + 1.j*im
    else:
        return _iuspline(xin, yin, k=k, ext=ext)(xout)

def _splinterp_Cwrapper(xout, xin, yin):
    """Uses gsl splines with a wrapper to interpolate real or complex data.
    Uses natural boundary conditions instead of not-a-knot boundary conditions
    like InterpolatedUnivariateSpline."""
    if len(xin) != len(yin):
        raise Exception('Expected x and y input lengths to match.')
    if np.iscomplexobj(yin):
        re = _splinterp_Cwrapper(xout, xin, np.real(yin))
        im = _splinterp_Cwrapper(xout, xin, np.imag(yin))
        return re + 1.j*im
    else:
        return spline_interp_Cwrapper.interpolate(xout, xin, yin)


class ParamDim(SimpleH5Object):
    """
    A helper class containing the information and functions for a single
    parameter space dimension
    """

    def __init__(self, name='', min_val=0, max_val=1, rtol=PARAM_NUDGE_TOL):
        """
        name: A descriptive name for this parameter dimension
        min_val: The minimum allowed value for this parameter
        max_val: The maximum allowed value for this parameter
        rtol: A relative tolerance for nudging parameters lying outside
             [min_val, max_val]. Scaled by (max_val - min_val) for an absolute
             tolerance. Useful to avoid machine precision issues.
        """
        super(ParamDim, self).__init__()

        tol = rtol * (max_val - min_val)

        if min_val + 2*tol > max_val:
            raise Exception("tol %s is too large for %s with range [%s, %s]"%(
                            tol, name, min_val, max_val))

        self.name = name
        self.min_val = min_val
        self.max_val = max_val
        self.tol = tol
        self.tol_max = max_val - tol
        self.tol_min = min_val + tol

    def __str__(self):
        return self.name

    def __repr__(self):
        return '%s: [%s, %s] with tol %s'%(self.name, self.min_val,
                                           self.max_val, self.tol)

    def nudge(self, x):
        """
        Returns a nudged version of x lying within [min_val+tol, max_val-tol].
        x can have any shape. Returns x if it already lies in the interval.
        x must already lie within [min_val - tol, max_val + tol].
        """
        xmax = np.max(x)
        xmin = np.min(x)

        if xmax > self.tol_max:
            if xmax > self.max_val + self.tol:
                raise Exception("The maximum allowed %s is %s, got %s"%(
                                self.name, self.max_val, xmax))
            x = np.amin([x, np.ones(np.shape(x)) * self.tol_max], axis=0)

        if xmin < self.tol_min:
            if xmin < self.min_val - self.tol:
                raise Exception("The maximum allowed %s is %s, got %s"%(
                                self.name, self.min_val, xmin))
            x = np.amax([x, np.ones(np.shape(x)) * self.tol_min], axis=0)

        return x


class ParamSpace(SimpleH5Object):
    """
    A helper class for all the parameter domain related information and
    functions.

    NOTE: While all surrogate model's define an instance of the ParamSpace class,
          not all models use it beyond storing information for the user's benefit.
        
          test_model_regression.py does query each model's ParamSpace instance
          so all models must include it. 

          Example: the NRHybSur3dq8{_CCE}'s  _sur_dimless object instantiates
          ParamSpace, but it doesn't play a role in this surrogate model's 
          evaluation. Rather, the parameter space mapping is handled by
          the NRHybSur3dq8Fit class in nodeFunction.py.

    """

    def __init__(self, name='', params=[]):
        """
        name: A descriptive name for this parameter space
        params: A list of ParamDim instances, one per parameter dimension
        """
        super(ParamSpace, self).__init__(['name', 'dim'], ['_params'])

        self.name = name
        self._params = H5ObjectList(params)
        self.dim = len(params)

    def __str__(self):
        return self.name

    def __repr__(self):
        return '%s: %s'%(self.name, [p.name for p in self._params])

    def param_names(self):
        """Returns the list of names for each parameter space dimension"""
        return [p.name for p in self._params]

    def min_vals(self):
        """Returns a list of minimum parameter space values"""
        return [p.min_val for p in self._params]

    def max_vals(self):
        """Returns a list of maximum parameter space values"""
        return [p.max_val for p in self._params]

    def nudge_params(self, x):
        """
        Nudges parameters lying slightly outside the valid domain to the
        boundary. x can be a single 1d parameter vector or a 2d vector with
        shape (n_params, self.dim).
        """

        xshape = np.shape(x)

        # It's convenient to be able to accept a float instead of a length-1
        # array for 1d parameter spaces.
        if len(xshape) == 0:
            x = np.array([x])
            xshape = np.shape(x)

        if len(xshape) == 1:
            if len(x) != self.dim:
                raise Exception("Parameter space has dimension %s, got %s."%(
                                self.dim, xshape))
            res = np.array([p.nudge(xi) for p, xi in zip(self._params, x)])

        elif len(xshape) == 2:
            if xshape[1] != self.dim:
                raise Exception("Expecting array with shape (n, %s), got %s"%(
                                self.dim, xshape))
            res = np.array([p.nudge(xi) for p, xi in zip(self._params, x.T)])

        else:
            raise Exception("x should be 1d or 2d, got shape {}".format(xshape))

        return res

    def h5_prepare_subs(self):
        """Setup dummy subordinates before loading them"""
        params = [ParamDim() for _ in range(self.dim)]
        self._params = H5ObjectList(params)


class _SingleFunctionSurrogate_NoChecks(SimpleH5Object):
    """
    A surrogate model for a single (real or complex) function on a 1d domain.
    Skips sanity/validity checks, so should be called with sanitized inputs.
    Use SingleFunctionSurrogate for actual surrogates of single functions.
    """

    def __init__(self, name=None, ei_basis=None, node_functions=[]):
        """
        name: A descriptive name for the function this surrogate models
        ei_basis: A basis with shape (n_nodes, len(domain)).
                  Interpolation is done via nodes.dot(ei_basis)
        node_functions: A list of evaluators for each node.
                        Each one takes a parameter space vector x and returns
                        the node evaluated at x.
        """
        super(_SingleFunctionSurrogate_NoChecks, self).__init__(
                data_keys=['name', 'ei_basis', 'n_nodes'],
                sub_keys=['node_functions'])

        self.name = name
        self.ei_basis = ei_basis
        self.n_nodes = len(node_functions)
        self.node_functions = H5ObjectList(node_functions)

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __call__(self, x):
        """
        Evaluates the surrogate at x, returning the result.
        """
        nodes = np.array([nf(x) for nf in self.node_functions])
        return nodes.dot(self.ei_basis)

    def h5_prepare_subs(self):
        """Setup NodeFunctions before loading them"""
        tmp_nodes = [NodeFunction() for _ in range(self.n_nodes)]
        self.node_functions = H5ObjectList(tmp_nodes)


class SingleFunctionSurrogate(_SingleFunctionSurrogate_NoChecks):
    """
    A surrogate model for a single (real or complex) function on a 1d domain.
    """

    def __init__(self, name=None, domain=None, param_space=None,
                 ei_basis=None, node_functions=[]):
        """
        name: A descriptive name for the function this surrogate models
        domain: A 1d array of the monotonically increasing domain
                (time/frequency) values
        param_space: A ParamSpace for this surrogate
        ei_basis: A basis with shape (n_nodes, len(domain)).
                  Interpolation is done via nodes.dot(ei_basis)
        node_functions: A list of evaluators for each node.
                        Each one takes a parameter space vector x and returns
                        the node evaluated at x.
        """
        super(SingleFunctionSurrogate, self).__init__(name=name,
                ei_basis=ei_basis, node_functions=node_functions)

        if domain is not None and np.min(np.diff(domain)) <= 0.0:
            raise Exception("domain should be monotonically increasing")

        self.domain = domain
        self.param_space = param_space

        self._h5_data_keys.append('domain')
        self._h5_subordinate_keys.append('param_space')

    def h5_prepare_subs(self):
        super(SingleFunctionSurrogate, self).h5_prepare_subs()
        self.param_space = ParamSpace()

    def __repr__(self):
        return '%s (%s)'%(self.name, self.param_space.name)

    def __call__(self, x, domain=None):
        """
        Evaluates the surrogate at x, returning the result.
        domain: An optional 1d array of domain values. If given, the result is
                evaluated at the domain values.
        """
        # Verify valid parameters and domain
        x = self.param_space.nudge_params(x)
        if domain is not None:
            if domain[0] < self.domain[0] or domain[-1] > self.domain[-1]:
                raise Exception("Domain must lie in [%s, %s]"%(
                                self.domain[0], self.domain[-1]))

        res = super(SingleFunctionSurrogate, self).__call__(x)

        if domain is not None:
            res = _splinterp(domain, self.domain, res)

        return res


class _ManyFunctionSurrogate_NoChecks(SimpleH5Object):
    """
    A container for surrogates sharing the same parameter space and domain.
    Skips checks, assuming they have already been performed.
    """

    def __init__(self, name='', single_function_components={},
                 many_function_components={}, combine_func='identity'):
        """
        name: A descriptive name for this surrogate.
        single_function_components: A dictionary of components, where the
                values are (ei_basis, node_functions) tuples.
        many_function_components: A dictionary of _ManyFunctionSurrogate_NoCheck
                components, where the values are
                (combine_func, single_function_components,
                 many_function_components) tuples.
        combine_func: A key of RECOMBINATION_FUNCS used to combine subordinate
                      results into a single result.
        """
        super(_ManyFunctionSurrogate_NoChecks, self).__init__(
                data_keys=['name', 'func_keys', 'sur_keys', 'combine_func'],
                sub_keys=['func_subs', 'sur_subs'])

        self.name = name
        self.combine_func = combine_func
        self.func_keys = list(single_function_components.keys())
        self.sur_keys = list(many_function_components.keys())
        if len(set(self.func_keys).intersection(self.sur_keys)) > 0:
            raise Exception("Component keys must be unique! Got %s, %s"%(
                            self.func_keys, self.sur_keys))

        func_sub_dict = {}
        for k, (ei, nf) in single_function_components.items(): # inefficient in py2
            func_sub_dict[k] = _SingleFunctionSurrogate_NoChecks(k, ei, nf)
        self.func_subs = H5ObjectDict(func_sub_dict)

        sur_sub_dict = {}
        for k, (cf, sfc, mfc) in many_function_components.items(): # inefficient in py2
            sur_sub_dict[k] = _ManyFunctionSurrogate_NoChecks(k, sfc, mfc, cf)
        self.sur_subs = H5ObjectDict(sur_sub_dict)

    def h5_prepare_subs(self):
        """
        Initialize subordinate surrogate class instances so they can
        load their own data from h5 files.
        """
        self.func_subs = H5ObjectDict({k: _SingleFunctionSurrogate_NoChecks()
                                       for k in self.func_keys})
        self.sur_subs = H5ObjectDict({k: _ManyFunctionSurrogate_NoChecks()
                                      for k in self.sur_keys})

    def __str__(self):
        return self.name

    def __call__(self, x):
        func_evals = {k: sur(x) for k, sur in self.func_subs.iteritems()} # inefficient in py2
        sur_evals = {k: sur(x) for k, sur in self.sur_subs.iteritems()} # inefficient in py2
        return RECOMBINATION_FUNCS[self.combine_func](func_evals, sur_evals)

    def _eval_func(self, x, key):
        return self.func_subs[key](x)

    def _eval_sur(self, x, key):
        return self.sur_subs[key](x)


class ManyFunctionSurrogate(_ManyFunctionSurrogate_NoChecks):
    """
    A container for surrogates sharing the same parameter space and domain.
    """

    def __init__(self, name='', domain=None, param_space=None,
                 single_function_components={}, many_function_components={},
                 combine_func='identity'):
        """
        name: A descriptive name for this surrogate.
        param_space: A ParamSpace for this surrogate.
        single_function_components: A dictionary of components, where the
                values are (ei_basis, node_functions) tuples.
        many_function_components: A dictionary of _ManyFunctionSurrogate_NoCheck
                components, where the values are
                (combine_func, single_function_components,
                 many_function_components) tuples.
        combine_func: A key of RECOMBINATION_FUNCS used to combine subordinate
                      results into a single result.
        """
        super(ManyFunctionSurrogate, self).__init__(
                name=name,
                single_function_components=single_function_components,
                many_function_components=many_function_components,
                combine_func=combine_func,
            )

        self.domain = domain
        self.param_space = param_space
        self._h5_data_keys.append('domain')
        self._h5_subordinate_keys.append('param_space')

    def h5_prepare_subs(self):
        super(ManyFunctionSurrogate, self).h5_prepare_subs()
        self.param_space = ParamSpace()

    def __call__(self, x):
        """
        Evaluates the surrogate at x, returning the result.
        """
        # Verify valid parameters
        x = self.param_space.nudge_params(x)
        return super(ManyFunctionSurrogate, self)(x)

class FastTensorSplineSurrogate(SimpleH5Object):
    """
    A special case of having a complex empirical interpolant combined with
    tensor splines for the real and imaginary parts of each empirical node,
    for each waveform mode. All tensor splines must use the same grid.
    Written for speed, minimizing python operations
    between tensor spline interpolations, which are done simultaneously for
    each mode to keep numpy busy. Obtained ~25ms evaluation time per waveform
    for 12 waveform modes. This was ~66ms when using the python class
    hierarchy, and using a separate call to numpy for each tensor spline
    interpolation. Note that similar C code written with gsl splines takes
    ~50ms, but should have room for optimization.
    """

    def __init__(self, name=None, domain=None, param_space=None,
                 knot_vecs=[], mode_data={}, modes=None):

        super(FastTensorSplineSurrogate, self).__init__(
                sub_keys=['param_space', 'ts_grid'],
                data_keys=['name', 'domain', 'ei', 'cre', 'cim', 'mode_list',
                           'mode_indices'])

        self.name = name
        self.domain = domain
        self.param_space = param_space
        if param_space is None:
            self.param_space = ParamSpace()
        if modes is None:
            modes = list(mode_data.keys())
        self.mode_list = modes
        self.mode_indices = {str(k): i for i, k in enumerate(modes)}
        self.ei = [mode_data[k][0] for k in modes]
        self.cre = [mode_data[k][1] for k in modes]
        self.cim = [mode_data[k][2] for k in modes]
        self.ts_grid = TensorSplineGrid(knot_vecs)

    def __call__(self, x, theta=None, phi=None, modes=None):
        """
        Return surrogate evaluation.
        Arguments:
            x : The intrinsic parameters (see self.param_space)
            theta/phi : polar and azimuthal angles of the direction of
                        gravitational wave emission. If given, sums up modes
                        and returns h_plus and h_cross (default returns modes)
            modes : A list of (ell, m) modes to be evaluated (default: all)
        Returns h:
            h : If theta and phi are None, h is a dictionary of waveform modes
                sampled at self.domain with (ell, m) keys.
                If theta and phi are given, h = h_plus - i * h_cross is a
                complex array given by the sum of the modes.
        """
        if (theta is None) != (phi is None):
            raise Exception("Either give theta and phi or neither")

        x = self.param_space.nudge_params(x)

        if modes is None:
            modes = self.mode_list

        h_modes = {}
        for k in modes:
            i = self.mode_indices[str(k)]

            h_eim = fast_complex_tensor_spline_eval(x,self.ts_grid,self.cre[i],self.cim[i])

            # Evaluate the empirical interpolant
            h_modes[k] = h_eim.dot(self.ei[i])

        if theta is not None:
            return _mode_sum(h_modes, theta, phi)

        return h_modes


class MultiModalSurrogate(ManyFunctionSurrogate):
    """
    A surrogate for multimodal waveforms, where each waveform mode has
    its own surrogate. Contains added functionality for evaluating on
    sphere.
    """

    def __init__(self, name=None, domain=None, param_space=None,
                 mode_data={}, mode_type='complex', modes=None):
        """
        name: A descriptive name for this surrogate.
        domain: A 1d array of the monotonically increasing domain
                (time/frequency) values
        param_space: A ParamSpace for this surrogate.
        mode_data: A dictionary of modes with (l, m) integer keys,
                   where the values are (ei_basis, node_functions) tuples.
        mode_type: Can be 'amp_phase', 're_im', or 'complex' depending on
                   how mode surrogates are built. If 'amp_phase' or 're_im',
                   mode_data values should instead be dictionaries with keys
                   ['amp', 'phase'] or ['re', 'im'] and
                   (ei_basis, node_functions) values.
        modes: A list of (ell, m) modes giving an ordering to mode_data.keys().
               If None, uses mode_data.keys().
        """

        if mode_type == 'complex':
            super(MultiModalSurrogate, self).__init__(name, domain, param_space,
                                                      mode_data, {}, 'identity')
        elif mode_type in ['amp_phase', 're_im']:
            mode_data = {k: (mode_type, v, {})
                         for k, v in mode_data.items()} # inefficient in py2
            super(MultiModalSurrogate, self).__init__(name, domain, param_space,
                                                      {}, mode_data, mode_type)
        else:
            raise ValueError("Invalid mode_type: %s"%(mode_type))

        self.mode_type = mode_type
        if modes is None:
            self.modes = list(mode_data.keys())
        else:
            self.modes = modes

        self._h5_data_keys.append('modes')
        self._h5_data_keys.append('mode_type')


    def __call__(self, x, theta=None, phi=None, modes=None):
        """
        Return surrogate evaluation.
        Arguments:
            x : The intrinsic parameters (see self.param_space)
            theta/phi : polar and azimuthal angles of the direction of
                        gravitational wave emission. If given, sums up modes
                        and returns h_plus and h_cross (default returns modes)
            modes : A list of (ell, m) modes to be evaluated (default: all)
        Returns h:
            h : If theta and phi are None, h is a dictionary of waveform modes
                sampled at self.domain with (ell, m) keys.
                If theta and phi are given, h = h_plus - i * h_cross is a
                complex array given by the sum of the modes.
        """
        if (theta is None) != (phi is None):
            raise Exception("Either give theta and phi or neither")

        x = self.param_space.nudge_params(x)

        if modes is None:
            modes = self.modes

        if self.mode_type == 'complex':
            h_modes = {k: self._eval_func(x, k) for k in modes}
        else:
            h_modes = {k: self._eval_sur(x, k) for k in modes}

        if theta is not None:
            return _mode_sum(h_modes, theta, phi)

        return h_modes


class AlignedSpinCoOrbitalFrameSurrogate(ManyFunctionSurrogate):
    """
    A surrogate for coorbital frame multimodal waveforms, where each waveform
    data piece has its own surrogate.

    The waveform data pieces are:
    Amplitude and phase of the (2,2) mode.
    Real and imaginary parts of coorbital frame waveform for other modes.
    """

    def __init__(self, name=None, domain=None, param_space=None, \
            phaseAlignIdx=None, TaylorT3_t_ref=None, \
            coorb_mode_data={(2, 2): {}}
            ):
        """
        name:               A descriptive name for this surrogate.

        domain:             A 1d array of the monotonically increasing time
                            values.

        param_space:        A ParamSpace for this surrogate.

        phaseAlignIdx:      This value should be loaded directly from the
            surrogate's h5 file. Index of domain at which the orbital phase is
            aligned. This is used when putting back the TaylorT3 contribution
            that was subtracted before modeling the phase.

        TaylorT3_t_ref:     This value should be loaded directly from the
            surrogate's h5 file. This is an arbitrary reference time used
            in the TaylorT3 contribution, but is fixed during the surrogate
            construction.

        coorb_mode_data: A dictionary of modes with (l, m) integer keys, where
            the values are themselves dictionaries containing the coorbital
            frame waveform for that mode. The coorbital frame is defined as:
            H_lm = h_lm*exp(i*m*phi_22/2), where h_lm is the inertial frame
            waveform and h_22 = A_22 * exp(-i phi_22). NOTE the minus sign.

            coorb_mode_data should be a dict with mode_key: mode_value pairs,
                where mode_value = AmpPhase_22 for mode_key = (2, 2)
                and mode_value = CoorbReIm_lm for other modes.

            Above, AmpPhase_22 and CoorbReIm_lm are themselved dictionaries:
                AmpPhase_22 = {'amp': Amp_22, 'phase': phi_22}
                CoorbReIm_lm = {'re': re_H_lm, 'im': im_H_lm}, where
                re_H_lm = Real(H_lm) and im_H_lm = Imag(H_lmc).
            Finally, all of Amp_22, phi_22, re_H_lm and im_H_lm should be
            (ei_basis, node_functions) tuples of that data piece.

            IMPORTANT NOTE: The phase of (2, 2) mode should be defined with
            a minus sign as shown above, this is opposite to what is done for
            MultiModalSurrogate.
        """

        # get list of modes, but move (2,2) mode to start of the list.
        # This is important because we need the phase of the 22 mode to
        # transform the other modes from coorbital frame to inertial frame.
        self.mode_list = list(coorb_mode_data.keys())
        mode22_idx = [i for i in range(len(self.mode_list)) \
            if self.mode_list[i] == tuple([2, 2])]
        if len(mode22_idx) != 1:
            raise Exception('Seems to have found multiple or no 22 mode!')
        mode22_idx = mode22_idx[0]

        # shift 22 mode to the first index
        self.mode_list.insert(0, self.mode_list.pop(mode22_idx))

        if self.mode_list[0] != tuple([2, 2]):
            raise Exception('Expected the first mode at this point to be the'\
                ' 22 mode.')
        # make sure shifting the 22 mode index did not delete or add a
        # mode by mistake
        if len(self.mode_list) != len(coorb_mode_data.keys()):
            raise Exception('Number of modes do not agree')

        self.mode_type = 'identity'
        many_function_components = {}
        for mode in self.mode_list:
            many_function_components[mode] = ('identity', \
                coorb_mode_data[mode], {})

        # required for TaylorT3
        self.phaseAlignIdx = phaseAlignIdx
        self.TaylorT3_t_ref = TaylorT3_t_ref
        self.TaylorT3_factor_without_eta = None

        super(AlignedSpinCoOrbitalFrameSurrogate, self).__init__(name,
                domain, param_space, {}, many_function_components,
                self.mode_type)

        self._h5_data_keys.append('mode_list')
        self._h5_data_keys.append('mode_type')
        self._h5_data_keys.append('phaseAlignIdx')
        self._h5_data_keys.append('TaylorT3_t_ref')

    def _search_omega(self, omega22, omega_val):
        """ Find closest index such that omega22[index] = omega_val
        """
        # find first index where omega22 > omega_val
        idx = np.where(omega22 > omega_val)[0][0]
        # if idx-1 is closer to omega_val, pick that instead
        if abs(omega22[idx-1] - omega_val) < abs(omega22[idx] - omega_val):
            idx -= 1
        return idx

    def _coorbital_to_inertial_frame(self, h_coorb, h_22, mode_list, dtM,
        timesM, fM_low, fM_ref, do_not_align):
        """ Transforms a dict from Coorbital frame to inertial frame.

            The surrogate data is sparsely sampled, so upsamples to time
            step dtM if given. This is done in the coorbital frame since
            the waveform is slowly varying in that frame.

            If fM_low is given, only part of the waveform where frequency of
            the (2, 2) mode is greater than fM_low is retained.

            if do_not_align = False:
                Aligns the 22 mode phase to be 0 at fM_ref. This means
                that at this reference frequency, the heavier BH is roughly on
                the +ve x axis and the lighter BH is on the -ve x axis.
            do_not_align should be True only when converting from pySurrogate
            format to gwsurrogate format as we may want to do some checks that
            the waveform has not been modified
        """

        Amp_22 = h_22[0]['amp']
        phi_22 = h_22[0]['phase']
        domain = np.copy(self.domain)

        # Get omega22_sparse, the angular frequency of the 22 mode, from the
        # sparse surrogate domain.
        # Use np.diff instead of np.gradient to match the LAL version
        omega22_sparse = np.append(np.diff(phi_22)/np.diff(domain), 0)

        # t=0 is at the waveform peak for the surrogate
        peak22Idx = np.argmin(np.abs(domain))
        omega22_peak = omega22_sparse[peak22Idx]
        # We ignore the part after the peak.  This way we avoid the noisy part
        # at late times, which can randomly be at frequency = fM_low.
        omega22_sparse = omega22_sparse[domain <= domain[peak22Idx]]

        # Get initIdx such that the initial (2, 2) mode frequency ~ fM_low.
        # We will make this more precise below.
        if fM_low != 0:
            omega_low = 2*np.pi*fM_low
            if omega_low < omega22_sparse[0]:
                raise ValueError('f_low is lower than the minimum allowed'
                    ' frequency')
            if omega_low > omega22_peak:
                raise ValueError('f_low is higher than the peak frequency')

            # Choose 5 indices less, to ensure omega_low is included
            initIdx = self._search_omega(omega22_sparse, omega_low) - 5
            # But if initIdx < 0, we are at the start of the surrogate data
            # so just choose 0
            if initIdx < 0:
                initIdx = 0

        else:
            # If fM_low is 0, we use the entire waveform
            initIdx = 0

            # But, if fM_low = 0 and timesM is given, the output of the
            # interpolant depends very slightly on the length of the sparse
            # data used to construct the interpolant. So, to achieve machine
            # precision equivalence between using dtM and timesM options, we
            # need to do the following: truncate before interpolation to the
            # same index as the dtM option would have used above (initIdx).
            # Using 6 rather than 5 because of the greater than condition.
            if timesM is not None:
                initIdx = np.where(domain > timesM[0])[0][0] - 6

        Amp_22 = Amp_22[initIdx:]
        phi_22 = phi_22[initIdx:]
        domain = domain[initIdx:]

        if timesM is not None:
            if timesM[-1] > domain[-1]:
                raise Exception("'times' includes times larger than the"
                    " maximum time value in domain.")
            if timesM[0] < domain[0]:
                raise Exception("'times' starts before start of domain. Try"
                    " increasing initial value of times or reducing f_low.")

        if dtM is None and timesM is None:
            # Use the sparse domain
            timesM = domain
            omega22 = omega22_sparse[initIdx:]
            do_interp = False
        else:
            ## Interpolate onto uniform-domain/timesM if needed
            do_interp = True
            if dtM is not None:
                t0 = domain[0]
                tf = domain[-1]
                num_times = int(np.ceil((tf - t0)/dtM));
                timesM = t0 + dtM*np.arange(num_times)
            else:
                if timesM[0] < domain[0] or timesM[-1] > domain[-1]:
                    raise Exception('Trying to evaluate at times outside the'
                        ' domain.')

            Amp_22 = _splinterp_Cwrapper(timesM, domain, Amp_22)
            phi_22 = _splinterp_Cwrapper(timesM, domain, phi_22)

            # now recompute omega22 with the dense data, but retain only data
            # upto the peak to avoid the noisy part
            omega22 = np.append(np.diff(phi_22)/np.diff(timesM), 0)
            omega22 = omega22[timesM <= 0]

            # Truncate data so that only freqs above omega_low are retained
            # If timesM are already given, we don't need to truncate data
            if dtM is not None:
                if fM_low != 0:
                    startIdx = self._search_omega(omega22, omega_low)
                else:
                    # If fM_low is 0, we use the entire waveform
                    startIdx = 0

                Amp_22 = Amp_22[startIdx:]
                phi_22 = phi_22[startIdx:]
                omega22 = omega22[startIdx:]
                timesM = timesM[startIdx:]


        # Get reference index where waveform needs to be aligned.
        if (abs(fM_ref-fM_low) < 1e-13) and (dtM is not None):
            # This means that the data is already truncated at fM_low,
            # so we just need the first index for fM_ref=fM_low
            refIdx = 0
        else:
            omega_ref = 2*np.pi*fM_ref
            if omega_ref > omega22_peak:
                raise ValueError('f_ref is higher than the peak frequency')

            refIdx = self._search_omega(omega22, omega_ref)


        # do_not_align should be True only when converting from pySurrogate
        # format to gwsurrogate format as we may want to do some checks that
        # the waveform has not been modified
        if not do_not_align:
            # Set orbital phase to 0 refIdx. Note that the Coorbital
            # frame data is not affected by this constant phase shift.

            # The orbital phase is obtained as phi_22/2, so this leaves a pi
            # ambiguity.  But the surrogate data is already aligned such that
            # the heavier BH is on the +ve x-axis at t=-1000M. See Sec.VI.A.4
            # of arxiv:1812.07865, the resolves the pi ambiguity. This means
            # that the after the realignment, the orbital phase at reference
            # frequency is 0.
            phi_22 += -phi_22[refIdx]

        h_dict = {}
        for mode in mode_list:
            if mode == tuple([2, 2]):
                h_dict[mode] = Amp_22 * np.exp(-1j*phi_22)
            else:
                l,m = mode
                h_coorb_lm = 0
                if 're' in h_coorb[mode][0].keys():
                    h_coorb_lm += h_coorb[mode][0]['re'] + 1j * 0
                if 'im' in h_coorb[mode][0].keys():
                    h_coorb_lm += 1j*h_coorb[mode][0]['im']

                h_coorb_lm = h_coorb_lm[initIdx:]
                if do_interp:
                    h_coorb_lm = _splinterp_Cwrapper(timesM,domain,h_coorb_lm)

                h_dict[mode] = h_coorb_lm * np.exp(-1j*m*phi_22/2.)

        return timesM, h_dict, None     # None is for dynamics

    def _set_TaylorT3_factor(self):
        """ Sets a term used in the 0 PN TaylorT3 phase. See Eq.43 of
        arxiv.1812.07865.
        """
        # Set only once
        if self.TaylorT3_factor_without_eta is None:
            # TaylorT3_t_ref is arbitrary. This is where the phase diverges,
            # so we choose it much after ringdown. This matches what was used
            # in the construction of the surrogate. See discussion near Eq.43
            # of arxiv.1812.07865
            theta_without_eta = ((self.TaylorT3_t_ref -self.domain)/5)**(-1./8)
            self.TaylorT3_factor_without_eta = -2./theta_without_eta**5

    def _TaylorT3_phase_22(self, x):
        """ 0 PN TaylorT3 phase. See Eq.43 of arxiv.1812.07865
        """

        q = x[0]
        eta = q/(1.+q)**2

        # 0PN TaylorT3 phase
        phi22_T3 = 1./eta**(3./8) * self.TaylorT3_factor_without_eta

        # Align at phaseAlignIdx
        phi22_T3 -= phi22_T3[self.phaseAlignIdx]

        return phi22_T3


    def __call__(self, x, fM_low=None, fM_ref=None, dtM=None,
            timesM=None, dfM=None, freqsM=None, mode_list=None, ellMax=None,
            precessing_opts=None, tidal_opts=None, par_dict=None,
            return_dynamics=False, do_not_align=False):
        """
    Return dimensionless surrogate modes.
    Arguments:
    x :             The intrinsic parameters EXCLUDING total Mass (see
                    self.param_space)

    fM_low :        Initial frequency of (2,2) mode in units of cycles/M.
                    If 0, will use the entire data of the surrogate.
                    Default None.

    fM_ref:         Frequency used to set the reference epoch at which
                    the reference frame is defined and the spins are specified.
                    See below for definition of the reference frame.
                    Default: None.

                    For time domain models, f_ref is used to determine a t_ref,
                    such that the frequency of the (2, 2) mode equals f_ref at
                    t=t_ref.

    dtM :           Uniform time step to use, in units of M. If None, the
                    returned time array will be the array used in the
                    construction of the surrogate, which can be nonuniformly
                    sampled.
                    Default None.

    timesM:         Time samples to evaluate the waveform at. Use either dtM or
                    timesM, not both.

    dfM :           This should always be None as for now we are assuming
                    a time domain model.

    freqsM:         Frequency samples to evaluate the waveform at. Use either
                    dfM or freqsM, not both.

    ellMax:         Maximum ell index for modes to include. All available m
                    indicies for each ell will be included automatically.
                    Default: None, in which case all available modes wll be
                    included.

    mode_list :     A list of (ell, m) modes to be evaluated.
                    Default None, which evaluates all avilable modes.
                    Will deduce the m<0 modes from m>0 modes.

    par_dict:       This should always be None for this model.

    do_not_align:   Ignore fM_ref and do not align the waveform. This should be
                    True only when converting from pySurrogate format to
                    gwsurrogate format as we may want to do some checks that
                    the waveform has not been modified.

    Returns
    timesM, h, dynamics:
        timesM : time array in units of M.
        h : A dictionary of waveform modes sampled at timesM with
            (ell, m) keys.
        dynamics: None, since this is a nonprecessing model.


    IMPORTANT NOTES:
    ===============

    The reference frame (or inertial frame) is defined as follows:
        The +ve z-axis is along the orbital angular momentum at the reference
        epoch. The separation vector from the lighter BH to the heavier BH at
        the reference epoch is along the +ve x-axis. The y-axis completes the
        right-handed triad. The reference epoch is set using f_ref.
        """

        if dfM is not None:
            raise ValueError('Expected dfM to be None for a Time domain model')
        if freqsM is not None:
            raise ValueError('Expected freqsM to be None for a Time domain'
                ' model')

        if mode_list is None:
            mode_list = self.mode_list
        if ellMax is not None:
            if ellMax > np.max(np.array(self.mode_list).T[0]):
                raise ValueError('ellMax is greater than max allowed ell.')
            include_modes = np.array(self.mode_list).T[0] <= ellMax
            mode_list = [self.mode_list[idx]
                    for idx in range(len(self.mode_list))
                    if include_modes[idx]]

        if par_dict is not None:
            raise ValueError('par_dict should be None for this model')

        # always evaluate the (2,2) mode, the other modes neeed this
        # for transformation from coorbital to inertial frame

        # At this stage the phase of the (2,2) mode is the residual after
        # removing the TaylorT3 part (see. Eq.44 of arxiv.1812.07865)
        h_22 = self._eval_sur(x, tuple([2, 2]))

        # Get the TaylorT3 part and add to get the actual phase
        self._set_TaylorT3_factor()
        h_22[0]['phase'] += self._TaylorT3_phase_22(x)

        h_coorb = {k: self._eval_sur(x, k) for k in mode_list \
                        if k != tuple([2,2])}

        return self._coorbital_to_inertial_frame(h_coorb, h_22, \
            mode_list, dtM, timesM, fM_low, fM_ref, do_not_align)

class AlignedSpinCoOrbitalFrameSurrogateTidal(AlignedSpinCoOrbitalFrameSurrogate):
    """
    A surrogate for coorbital frame multimodal waveforms, where each waveform
    data piece has its own surrogate.

    The waveform data pieces are:
    Amplitude and phase of the (2,2) mode.
    Real and imaginary parts of coorbital frame waveform for other modes.

    This generates tidal inspiral waveforms by taking the surrogate output tuned
    to BBH results and incorporates the PN tidal corrections according to the
    tidal splicing method

    NOTE: This returns the waveform only during the inspiral portion of the
    binary's evolution, where the PN expansion is still valid; additional work
    will need to be done in order to complete the merger/ringdown portion of the
    waveform

    NOTE: The waveform is output with the time set so that t=0 corresponds to
    the peak of the waveform for the BBH waveform from the underlying surrogate,
    and NOT the peak of the tidally spliced waveform
    """

    def _coorbital_to_inertial_frame(self, h_coorb, h_22, mode_list, dtM,
        timesM, fM_low, fM_ref, do_not_align, x):
        """ Transforms a dict from Coorbital frame to inertial frame.

            The surrogate data is sparsely sampled, so upsamples to time
            step dtM if given. This is done in the coorbital frame since
            the waveform is slowly varying in that frame.

            If fM_low must be specified. The option of fM_low == 0 has been
            turned off for this model because of its excessive computational
            cost to evaluate

            if do_not_align = False:
                Aligns the 22 mode phase to be 0 at fM_ref. This means
                that at this reference frequency, the heavier BH is roughly on
                the +ve x axis and the lighter BH is on the -ve x axis.
            do_not_align should be True only when converting from pySurrogate
            format to gwsurrogate format as we may want to do some checks that
            the waveform has not been modified
        """

        Amp_22 = h_22[0]['amp']
        phi_22 = h_22[0]['phase']
        domain = np.copy(self.domain)

        # Get omega22_sparse, the angular frequency of the 22 mode, from the
        # sparse surrogate domain.
        # Use np.gradient
        omega22_sparse = np.gradient(phi_22, domain)

        # t=0 is at the waveform peak for the surrogate
        peak22Idx = np.argmin(np.abs(domain))
        omega22_peak = omega22_sparse[peak22Idx]
        # We ignore the part after the peak.  This way we avoid the noisy part
        # at late times, which can randomly be at frequency = fM_low.
        omega22_sparse = omega22_sparse[domain <= domain[peak22Idx]]

        # Get initIdx such that the initial (2, 2) mode frequency ~ fM_low.
        # We will make this more precise below.
        if fM_low != 0:
            omega_low = 2*np.pi*fM_low
            if omega_low < omega22_sparse[0]:
                raise ValueError('f_low is lower than the minimum allowed'
                    ' frequency')
            if omega_low > omega22_peak:
                raise ValueError('f_low is higher than the peak frequency')

            # Choose 5 indices less, to ensure omega_low is included
            initIdx = self._search_omega(omega22_sparse, omega_low) - 5
            # But if initIdx < 0, we are at the start of the surrogate data
            # so just choose 0
            if initIdx < 0:
                initIdx = 0
        else:
            raise ValueError("The option of setting 'fM_low' to 0 is turned off"
                    " for this model; must specifiy a non-zero 'fM_low'")
            ## If fM_low is 0, we use the entire waveform where frequency is
            ## monotonic, uncomment if want to allow this option
            #freq_orbital = np.gradient(phi_22[:peak22Idx], domain[:peak22Idx])
            #if np.min(np.diff(freq_orbital))<=0:
            #  initIdx = len(freq_orbital)-np.argmin((np.diff(freq_orbital)>np.zeros(len(freq_orbital)-1))[::-1])-1
            #else:
            #  initIdx = 0

        Amp_22 = Amp_22[initIdx:peak22Idx]
        phi_22 = phi_22[initIdx:peak22Idx]
        domain = domain[initIdx:peak22Idx]
        v_domain = np.power(np.abs(np.gradient(phi_22, domain))/2,1./3.)
        if(np.min(np.diff(v_domain))<0):
            raise ValueError('frequency is not monotonic over the entire'
                ' considered here')

        if timesM is not None:
            # This check is performed after the tidal terms computed
            #if timesM[-1] > domain[-1]:
            #    raise Exception("'times' includes times larger than the"
            #        " maximum time value in domain.")
            if timesM[0] < domain[0]:
                raise Exception("'times' starts before start of domain. Try"
                    " increasing initial value of times or reducing f_low.")

        # For tidal splicing, always want to interpolate first to a dense domain
        # in order to compute an accurate orbital frequency for the PN equations
        # then interpolated to the desired times afterwards

        if dtM is None and timesM is None:
            raise ValueError("For this model, must specify either the 'dtM' or"
                " 'timesM' option")
        else:
            ## Interpolate onto uniform domain
            ## WARNING -- if the the time points are not sampled densely enough
            ## here, there is a potential for error due to inaccurate orbital
            ## freq being used for the PN tidal equations
            if dtM is not None:
                t0 = domain[0]
                tf = domain[-1]
                num_times = int(np.ceil((tf - t0)/dtM));
                timesM_tmp = t0 + dtM*np.arange(num_times)
            else:
                # Because the spliced waveform is shifted so the final time
                # is the peak of the final waveform, we must ensure the check
                # here is performed similarly
                if timesM[0] < (domain[0]-domain[-1]) or timesM[-1] > 0:
                    raise Exception('Trying to evaluate at times outside the'
                        ' domain.')
                min_dt = np.min(np.diff(timesM))
                t0 = domain[0] #timesM[0] - min_dt
                tf = domain[-1]
                num_times = int(np.ceil((tf - t0)/min_dt));
                timesM_tmp = t0 + min_dt*np.arange(num_times)

            Amp_22 = _splinterp_Cwrapper(timesM_tmp, domain, Amp_22)
            phi_22 = _splinterp_Cwrapper(timesM_tmp, domain, phi_22)

            # now recompute omega22 with the dense data, but retain only data
            # upto the peak to avoid the noisy part
            omega22 = np.gradient(phi_22, timesM_tmp)

            #omega22 = omega22[timesM_tmp <= timesM_tmp[np.argmax(Amp_22)]]

            # Truncate data so that only freqs above omega_low are retained
            # If timesM are already given, we don't need to truncate data
            if dtM is not None:
                if fM_low != 0:
                    startIdx = max(np.argmin(np.abs(omega22 - omega_low)) - 4,0)
                else:
                    raise ValueError("The option of setting 'fM_low' to 0"
                            " is turned off for this model; must specifiy a"
                            " non-zero 'fM_low'")
                    ## If fM_low is 0, we use the entire waveform that is monotonic
                    #startIdx = 0
                    #if np.min(np.diff(omega22))<=0:
                    #  startIdx = len(omega22)-np.argmin((np.diff(omega22)>np.zeros(len(omega22)-1))[::-1])-1
                    ## Because the splicing changes the frequencies slightly, to
                    ## ensure we have wiggle room for interpolation later, buffer
                    ## the altered initial frequency of the spliced waveform so
                    ## it is not less than the initial frequency of v_domain
                    #gap = int((domain[5]-domain[0])/(timesM_tmp[1]-timesM_tmp[0]))
                    #if startIdx<gap:
                    #  startIdx=gap

                Amp_22 = Amp_22[startIdx:]
                phi_22 = phi_22[startIdx:]
                omega22 = omega22[startIdx:]
                timesM_tmp = timesM_tmp[startIdx:]

        freq_orbital = np.abs(omega22)/2
        v = np.power(freq_orbital,1./3.)

        # Setup all of the tidal parameters
        # Use universal relations to compute parameters beyond the quad love num
        # NOTE: omega2AB and omega3AB are stored as M*omega{2,3}{A,B}, to use the
        #   dimensionless value set the total mass of the system and that the
        #   universal relations return M{A,B}*omega{2,3}{A,B}
        # Aqm is the dimensionless quadrupole moment, however for splicing the 2PN BBH
        #   (v^4) term, the qm of a BBH must be subtracted off (Aqm_BH = 1; see
        #   arXiv:gr-qc/9709032 just below eqn 8), which will be done in the tidal
        #   function itself
        # If the NS is spinning, the effective driving frequency that the NS sees,
        #   from its own reference frame, is shifted according to the NS spin by the
        #   dimensionless rotation = M omega_spin =  (M / m_NS) * chi_NS / \bar{I}
        # WARNING: This effect has been turned off (omega_spin = 0) for NS anti-
        #   aligned spins, b/c the resonance peak is shifted to early enough in the
        #   inspiral that the approximation being used to model it might be breaking
        #   down by the end of the late inspiral (only supposed to be good up until
        #   shortly after resonance)
        qqq = x[0]; chiAz = x[1]; chiBz = x[2]; lambda2A = x[3]; lambda2B = x[4]
        XA = qqq/(1.+qqq); XB = 1.-XA
        omega2A = lambda3A = omega3A = AqmA = 0.
        omega2B = lambda3B = omega3B = AqmB = 0.
        omegaSpinA = omegaSpinB = 0.
        ell2Adyn = ell2Adiss = ell2Bdyn = ell2Bdiss = np.zeros(len(timesM_tmp))
        ell3Adyn = ell3Bdyn = np.zeros(len(timesM_tmp))
        if(lambda2A>0):
            IbarA     = UniversalRelationLambda2ToI(lambda2A)
            omegaSpinA = max(chiAz,0) / IbarA / XA
            omega2A   = UniversalRelationLambda2ToOmega2(lambda2A)/XA
            lambda3A  = UniversalRelationLambda2ToLambda3(lambda2A)
            omega3A   = UniversalRelationLambda3ToOmega3(lambda3A)/XA
            AqmA      = UniversalRelationLambda2ToAqm(lambda2A)
            ell2Adyn  = EffectiveDeformabilityFromDynamicalTides \
                        (np.abs(freq_orbital-omegaSpinA),omega2A,2,qqq)
            ell3Adyn  = EffectiveDeformabilityFromDynamicalTides \
                        (np.abs(freq_orbital-omegaSpinA),omega3A,3,qqq)
        if(lambda2B>0):
            IbarB     = UniversalRelationLambda2ToI(lambda2B)
            omegaSpinB = max(chiBz,0) / IbarB / XB
            omega2B   = UniversalRelationLambda2ToOmega2(lambda2B)/XB
            lambda3B  = UniversalRelationLambda2ToLambda3(lambda2B)
            omega3B   = UniversalRelationLambda3ToOmega3(lambda3B)/XB
            AqmB      = UniversalRelationLambda2ToAqm(lambda2B)
            ell2Bdyn  = EffectiveDeformabilityFromDynamicalTides \
                        (np.abs(freq_orbital-omegaSpinB),omega2B,2,qqq)
            ell3Bdyn  = EffectiveDeformabilityFromDynamicalTides \
                        (np.abs(freq_orbital-omegaSpinB),omega3B,3,qqq)

        dt_tid, dp_tid = PNT2Tidal(v, qqq, lambda2A*ell2Adyn, \
                lambda3A*ell3Adyn, AqmA, chiAz, lambda2B*ell2Bdyn, \
                lambda3B*ell3Bdyn, AqmB, chiBz, order=5)

        timesM_tmp += dt_tid - dt_tid[0]

        # Limit the waveform to the last time in the array that is increasing
        find = np.argmin(np.diff(timesM_tmp)>0)
        if(find == 0):
            find = len(timesM_tmp)

        timesM_tmp = timesM_tmp[:find]

        # There is a small region of parameter space where the interpolation
        # behaves poorly at very late times due to oddly shaped steps, so we
        # need to check the final handful of steps for that and truncate as
        # needed to avoid interpolation failures
        numcheck = 500
        factorLimit = 2.
        tdiff = np.diff(timesM_tmp[-numcheck-1:])
        for i in range(len(tdiff)-1):
          if ((tdiff[i]>tdiff[i+1]*factorLimit) or (tdiff[i]<tdiff[i+1]/factorLimit)):
            find = len(timesM_tmp)-numcheck+i-1
            timesM_tmp = timesM_tmp[:find]
            break

        timesM_tmp -= timesM_tmp[-1]
        phi_22 = phi_22[:find] + 2.*(dp_tid[:find] - dp_tid[0])

        # Reinterpolate to the final time grid
        if dtM is not None:
            t0 = timesM_tmp[0]
            tf = timesM_tmp[-1]
            num_times = int(np.ceil((tf - t0)/dtM));
            timesM = t0 + dtM*np.arange(num_times)
            timesM -= timesM[-1] # Ensure peak amplitude at t=0
        else:
            if timesM[-1] > timesM_tmp[-1]:
                raise Exception("'times' includes times larger than the"
                    " maximum time value in domain after splicing. (Remember"
                    " that tidal effects cause the binary to merger earlier)")
            if timesM[0] < timesM_tmp[0]:
                raise Exception("'times' includes times smaller than the"
                    " initial time value in domain after splicing. (Remember"
                    " that tidal effects cause the binary to merger earlier)")

        # Find the 'v' corresponding to the final time array, then perform the
        # interpolation in the 'v' domain as that is where most of the PN
        # quantities are defined
        v_uniform = _splinterp_Cwrapper(timesM, timesM_tmp, v[:find])

        Amp_22 = _splinterp_Cwrapper(v_uniform, v[:find], Amp_22[:find])
        phi_22 = _splinterp_Cwrapper(v_uniform, v[:find], phi_22)
        freq_orbital = np.power(v_uniform,3.)

        # Dynamical Tidal deformability stuff on final array for strain
        # amplitudes
        ell2Adyn = ell2Adiss = ell2Bdyn = ell2Bdiss = np.zeros(len(timesM))
        if(lambda2A>0):
          ell2Adyn  = EffectiveDeformabilityFromDynamicalTides \
                      (np.abs(freq_orbital-omegaSpinA),omega2A,2,qqq)
          ell2Adiss = EffectiveDissipativeDynamicalTides \
                      (np.abs(freq_orbital-omegaSpinA),ell2Adyn,omega2A,XA)
        if(lambda2B>0):
          ell2Bdyn  = EffectiveDeformabilityFromDynamicalTides \
                      (np.abs(freq_orbital-omegaSpinB),omega2B,2,qqq)
          ell2Bdiss = EffectiveDissipativeDynamicalTides \
                      (np.abs(freq_orbital-omegaSpinB),ell2Bdyn,omega2B,XB)

        # Get reference index where waveform needs to be aligned.
        if (abs(fM_ref-fM_low) < 1e-13) and (dtM is not None):
            # This means that the data is already truncated at fM_low,
            # so we just need the first index for fM_ref=fM_low
            refIdx = 0
        else:
            omega_ref = 2*np.pi*fM_ref
            if omega_ref > omega22_peak:
                raise ValueError('f_ref is higher than the peak frequency')

            refIdx = np.argmin(np.abs(2.*freq_orbital - omega_ref))


        # do_not_align should be True only when converting from pySurrogate
        # format to gwsurrogate format as we may want to do some checks that
        # the waveform has not been modified
        if not do_not_align:
            # Set orbital phase to 0 refIdx. Note that the Coorbital
            # frame data is not affected by this constant phase shift.

            # The orbital phase is obtained as phi_22/2, so this leaves a pi
            # ambiguity.  But the surrogate data is already aligned such that
            # the heavier BH is on the +ve x-axis at t=-1000M. See Sec.VI.A.4
            # of arxiv:1812.07865, the resolves the pi ambiguity. This means
            # that the after the realignment, the orbital phase at reference
            # frequency is 0.
            phi_22 += -phi_22[refIdx]


        h_dict = {}
        for mode in mode_list:
            if mode == tuple([2, 2]):
                h_dict[mode] = (Amp_22+StrainTidalEnhancementFactor(2,2, \
                      qqq,(lambda2A*ell2Adiss),(lambda2B*ell2Bdiss),v_uniform)) \
                      * np.exp(-1j*phi_22)
            else:
                l,m = mode
                h_coorb_lm = 0
                if 're' in h_coorb[mode][0].keys():
                    h_coorb_lm += h_coorb[mode][0]['re'] + 1j * 0
                if 'im' in h_coorb[mode][0].keys():
                    h_coorb_lm += 1j*h_coorb[mode][0]['im']

                h_coorb_lm = h_coorb_lm[initIdx:peak22Idx]

                h_coorb_lm = _splinterp_Cwrapper(v_uniform,v_domain,h_coorb_lm)

                h_coorb_lm_amp = np.abs(h_coorb_lm)
                h_coorb_lm_phase = np.unwrap(np.angle(h_coorb_lm))
                h_coorb_lm_tid = StrainTidalEnhancementFactor(l,m,qqq, \
                        (lambda2A*ell2Adiss),(lambda2B*ell2Bdiss),v_uniform)
                h_dict[mode] = (h_coorb_lm_amp + h_coorb_lm_tid) \
                        * np.exp((-1j*m/2.)*phi_22+1j*h_coorb_lm_phase)

        return timesM, h_dict, None     # None is for dynamics

    def __call__(self, x, fM_low=None, fM_ref=None, dtM=None,
        timesM=None, dfM=None, freqsM=None, mode_list=None, ellMax=None,
        precessing_opts=None, tidal_opts=None, par_dict=None,
        do_not_align=False):
        """
    Return dimensionless surrogate modes.
    Arguments:
    x :             The intrinsic parameters EXCLUDING total Mass (see
                    self.param_space)

    fM_low :        Initial frequency of (2,2) mode in units of cycles/M.
                    If 0, will use the entire data of the surrogate.
                    Default None.

    fM_ref:         Frequency used to set the reference epoch at which
                    the reference frame is defined and the spins are specified.
                    See below for definition of the reference frame.
                    Default: None.

                    For time domain models, f_ref is used to determine a t_ref,
                    such that the frequency of the (2, 2) mode equals f_ref at
                    t=t_ref.

    dtM :           Uniform time step to use, in units of M. If None, the
                    returned time array will be the array used in the
                    construction of the surrogate, which can be nonuniformly
                    sampled.
                    Default None.

    timesM:         Time samples to evaluate the waveform at. Use either dtM or
                    timesM, not both.

    dfM :           This should always be None as for now we are assuming
                    a time domain model.

    freqsM:         Frequency samples to evaluate the waveform at. Use either
                    dfM or freqsM, not both.

    ellMax:         Maximum ell index for modes to include. All available m
                    indicies for each ell will be included automatically.
                    Default: None, in which case all available modes wll be
                    included.

    mode_list :     A list of (ell, m) modes to be evaluated.
                    Default None, which evaluates all avilable modes.
                    Will deduce the m<0 modes from m>0 modes.

    par_dict:       This should always be None for this model.

    do_not_align:   Ignore fM_ref and do not align the waveform. This should be
                    True only when converting from pySurrogate format to
                    gwsurrogate format as we may want to do some checks that
                    the waveform has not been modified.

    Returns
    timesM, h, dynamics:
        timesM : time array in units of M.
        h : A dictionary of waveform modes sampled at timesM with
            (ell, m) keys.
        dynamics: None, since this is a nonprecessing model.


    IMPORTANT NOTES:
    ===============

    The reference frame (or inertial frame) is defined as follows:
        The +ve z-axis is along the orbital angular momentum at the reference
        epoch. The separation vector from the lighter BH to the heavier BH at
        the reference epoch is along the +ve x-axis. The y-axis completes the
        right-handed triad. The reference epoch is set using f_ref.
        """

        if par_dict is not None:
            raise ValueError('Expected par_dict to be None.')
        if dfM is not None:
            raise ValueError('Expected dfM to be None for a Time domain model')
        if freqsM is not None:
            raise ValueError('Expected freqsM to be None for a Time domain'
                ' model')

        if mode_list is None:
            mode_list = self.mode_list
        if ellMax is not None:
            if ellMax > np.max(np.array(self.mode_list).T[0]):
                raise ValueError('ellMax is greater than max allowed ell.')
            include_modes = np.array(self.mode_list).T[0] <= ellMax
            mode_list = [self.mode_list[idx]
                    for idx in range(len(self.mode_list))
                    if include_modes[idx]]

        # The last to parameters are the tidal parameters and are not a part of
        # the base surrogate model
        x_sur = x[:-2]

        # always evaluate the (2,2) mode, the other modes neeed this
        # for transformation from coorbital to inertial frame

        # At this stage the phase of the (2,2) mode is the residual after
        # removing the TaylorT3 part (see. Eq.44 of arxiv.1812.07865)
        h_22 = self._eval_sur(x_sur, tuple([2, 2]))

        # Get the TaylorT3 part and add to get the actual phase
        self._set_TaylorT3_factor()
        h_22[0]['phase'] += self._TaylorT3_phase_22(x_sur)

        h_coorb = {k: self._eval_sur(x_sur, k) for k in mode_list \
                        if k != tuple([2,2])}

        return self._coorbital_to_inertial_frame(h_coorb, h_22, \
            mode_list, dtM, timesM, fM_low, fM_ref, do_not_align, x)



class SpEC_nonspinning_q10_surrogate(MultiModalSurrogate):
    """A special class for the SpEC nonspinning surrogate"""

    skip_2_0_mode = True

    def __call__(self, x, theta=None, phi=None, modes=None,
                 fake_neg_modes=True):
        """
        Return surrogate evaluation.
        Arguments:
            x : The intrinsic parameters (see self.param_space)
            theta/phi : polar and azimuthal angles of the direction of
                        gravitational wave emission. If given, sums up modes
                        and returns h_plus and h_cross (default returns modes)
            modes : A list of (ell, m) modes to be evaluated (default: all)
            fake_neg_modes: Deduce (ell, -m) modes from (ell, m) modes for m>0.
        Returns h:
            h : If theta and phi are None, h is a dictionary of waveform modes
                sampled at self.domain with (ell, m) keys.
                If theta and phi are given, h = h_plus - i * h_cross is a
                complex array given by the sum of the modes.
        """
        if modes is None:
            modes = self.modes

        if self.skip_2_0_mode:
            # removed by 2to3 tool
            #modes = filter(lambda mode: mode != (2, 0), modes)
            modes = [mode for mode in modes if mode != (2, 0)]

        h_modes = super(SpEC_nonspinning_q10_surrogate, self).__call__(
                x, modes=modes)

        if fake_neg_modes:
            new_modes = {}
            for (ell, m), h in h_modes.items(): # inefficient in py2
                if m > 0:
                    new_modes[(ell, -m)] = np.power(-1, ell) * h.conjugate()
            for k, v in new_modes.items(): # inefficient in py2
                h_modes[k] = v

        if theta is not None:
            return _mode_sum(h_modes, theta, phi)

        return h_modes



"""
Test every gwsurrogate model.

Each model should test the following at a small handful of parameter values:

* all modes 
* summation of all modes
* each of the two above, for dimensionless and physical waveforms

Running this script as a regression test (with pytest) will download
regression data from Dropbox. 

Regression data should NOT be generated locally unless a new regression
file is being created to upload to Dropbox. In this case do

>>> python test_model_regression.py

from the folder test.

NOTES
=====
(1) Single precision regression data

Waveform regression data is saved with single precision in order to,

     (i) reduce the size of the regression file and
     (ii) allow h5diff to not fail due to round-off error (Still fails! switched comparison to np.allclose)

this is OK because the models themselves are not accurate to better than single precision.


(2) Dynamics surrogate not (directly) tested

No regression on dynamics surrogate output. This is probably OK since coorb
full surrogate uses dynamics output.

(3) As of 6/2023, attempt to log versions of some key packages that are known
    to impact regression data 


    (i) NRHybSur3dq8_CCE
            sklearn 1.2.2
            numpy   1.24.3
            GSL     2.4
            python  3.11
"""


from __future__ import division
import numpy as np
import gwsurrogate as gws
from gwsurrogate.new import surrogate 
import h5py, os, subprocess, time, warnings


import hashlib

def md5(fname):
  """ Compute has from file. code taken from 
  https://stackoverflow.com/questions/3431825/generating-an-md5-checksum-of-a-file"""
  hash_md5 = hashlib.md5()
  with open(fname, "rb") as f:
    for chunk in iter(lambda: f.read(4096), b""):
      hash_md5.update(chunk)
  return hash_md5.hexdigest()

# set global tolerances for floating point comparisons. 
# From documentation on np.testing.assert_allclose
# the comparison is
#
#     absolute(`a` - `b`) <= (`atol` + `rtol` * absolute(`b`))
#
# Setting atol=0 means we are only testing relative errors
#
atol = 0.0
# why a high tolerance? For some reason, a high tolerance is needed when 
# comparing to regression data on different machines
# TODO: explore the origin of these large discrepancies 
#       note that regression data model_regression_data.h5 is saved in single precision (see above)
#       largest relative errors seem to be post-merger
#       only seems to affect models that use gpr fits and/or gsl calls
rtol_default           = 1.e-11
rtols = {'NRHybSur3dq8':  2.e-5,
         'NRHybSur2dq15': 2.e-5,
         'NRHybSur3dq8_CCE': 7.e-5,  # higher modes for this GPR-fit model require a bit higher tolerance
         'NRHybSur3dq8Tidal': 3.e-4,
         'SpEC_q1_10_NoSpin_linear_alt': 3.e-8, # needed for (8,7) mode to pass. Other modes pass with 1e-11 tolerance
         }

# TODO: new and old surrogate interfaces should be similar enough to avoid
#       model-specific cases like below

# Old surrogate interface
surrogate_old_interface = ["SpEC_q1_10_NoSpin","EOBNRv2_tutorial","EOBNRv2","SpEC_q1_10_NoSpin_linear","EMRISur1dq1e4","BHPTNRSur1dq1e4"]

# news loader class
surrogate_loader_interface = ["NRHybSur3dq8","NRHybSur3dq8Tidal","NRSur7dq4","NRHybSur2dq15","NRHybSur3dq8_CCE"]

# Most models are randomly sampled, but in some cases its useful to provide 
# test points to activate specific code branches. This is done by mapping 
# a model (named in the surrogate catalog) to a function that will return 
# 3 unique parameter points.
#
# Each sampler should return a list [q, chiA, chiB] and a dictionary tidOpts
# and a dictionary precesingOpts
model_sampler = {}

def NRHybSur3dq8Tidal_samples(i):
  """ sample points for the NRHybSur3dq8Tidal model
  the ith sample point to evaluate the model at
  samples are returned as [q, chiA, chiB], tidOpts """

  assert i in [0,1,2]

  if i==0:
    return [1.2, [0.,0.,.1], [0.,0.,.1]], {'Lambda1': 1000.0, 'Lambda2': 4000.0}, None
  elif i==1:
    return [1.2, [0.,0.,.4], [0.,0.,-.4]], {'Lambda1': 0.0, 'Lambda2': 9000.0}, None
  elif i==2:
    return [1.2, [0.,0.,.0], [0.,0.,.0]], {'Lambda1': 0.0, 'Lambda2': 0.0}, None

model_sampler["NRHybSur3dq8Tidal"] = NRHybSur3dq8Tidal_samples


def NRSur7dq4_samples(i):
  """ sample points for the NRSur7dq4 model
  the ith sample point to evaluate the model at
  samples are returned as [q, chiA, chiB], precessingOpts """

  assert i in [0,1,2]

  if i==0:
    chiA = [-0.2, 0.4, 0.1]
    chiB = [-0.5, 0.2, -0.4]
    precessing_opts = {'init_quat': [1,0,0,0],
                       'return_dynamics': True,
                       'init_orbphase': 0.0}
                       #'use_lalsimulation_conventions': True}
    return [2., chiA, chiB], None, precessing_opts
  elif i==1:
    chiA = [-0.2, 0.4, 0.1]
    chiB = [-0.5, 0.2, -0.4]
    precessing_opts = {'init_quat': [1,0,0,0],
                       'return_dynamics': True,
                       'init_orbphase': 1.0}
                       #'use_lalsimulation_conventions': False}
    return [3., chiA, chiB], None, precessing_opts
  elif i==2:
    chiA = [-0.2, 0.4, 0.1]
    chiB = [-0.5, 0.2, -0.4]
    precessing_opts = {'init_quat': [1,0,0,0],
                       'return_dynamics': True,
                       'init_orbphase': 0.0}
                       #'use_lalsimulation_conventions': True}
    return [5., chiA, chiB], None, precessing_opts

model_sampler["NRSur7dq4"] = NRSur7dq4_samples

def BHPTNRSur1dq1e4_samples(i):
  """ sample points for the BHPTNRSur1dq1e4 model """

  assert i in [0,1,2]

  if i==0:
    return [3.0, [0,0,0],[0,0,0]], None, None
  elif i==1:
    return [100.0, [0,0,0],[0,0,0]], None, None
  elif i==2:
    return [10000.0, [0,0,0],[0,0,0]], None, None

model_sampler["BHPTNRSur1dq1e4"] = BHPTNRSur1dq1e4_samples


def NRHybSur2dq15_samples(i):
  """ sample points for the NRHybSur2dq15 model """

  assert i in [0,1,2]

  if i==0:
    return [2.0, [0,0,-.4],[0,0,0]], None, None
  elif i==1:
    return [11.0, [0,0,.7],[0,0,0]], None, None
  elif i==2:
    return [18.0, [0,0,.4],[0,0,0]], None, None

model_sampler["NRHybSur2dq15"] = NRHybSur2dq15_samples


def NRHybSur3dq8_CCE_samples(i):
  """ sample points for the NRHybSur3dq8_CCE model """
                       
  assert i in [0,1,2]  
    
  if i==0:
    return [2.0, [0,0,-.4],[0,0,0.5]], None, None
  elif i==1:
    return [7.0, [0,0,.7],[0,0,-0.8]], None, None
  elif i==2:           
    return [10.0, [0,0,0],[0,0,0.2]], None, None
                       
model_sampler["NRHybSur3dq8_CCE"] = NRHybSur3dq8_CCE_samples


def flatten_params(x):
  """ Convert [q, chiA, chiB] to [q, chiAx, chiAy, chiAz, chiBx, chiBy, chiBz].

  This function is only used when writting samples from specific model samplers
  to HDF5 file. """

  return [x[0],x[1][0],x[1][1],x[1][2],x[2][0],x[2][1],x[2][2]]


def test_model_regression(generate_regression_data=False):
  """ If generate_regression_data = True, this script will generate 
  an hdf5 file to diff against. No regression will be done.

  If generate_regression_data = False, this script will compare 
  model evaluations to the hdf5 file produced when True. In a typical use case,
  this regression file will be downloaded. """

  if generate_regression_data:
    h5_file = "model_regression_data_new.h5"
    print("Generating regression data file... Make sure this step is done BEFORE making any code changes!\n")
    print(os.path.exists(h5_file))
    if os.path.exists(h5_file):
      raise RuntimeError("Refusing to overwrite a regression file!")
  else:
    h5_file = "test/comparison_data.h5" # assumes pytest runs from project-level folder
    try: # try importing data. If it doesn't exist, download it
      fp_regression = h5py.File("test/model_regression_data.h5",'r') 
    except IOError:
      print("Downloading regression data...")
      # Old file (10/2022): https://www.dropbox.com/s/vxqsr7fjoffxm5w/model_regression_data.h5
      # w/o CCE file (4/2023): https://www.dropbox.com/s/copq3gx9a45ewy6/model_regression_data-OLD-CCE_delete-12-13-2023.h5
      os.system('wget -O test/model_regression_data.h5 https://www.dropbox.com/scl/fi/ajhhrku5odfy43vsdenu1/model_regression_data.h5?rlkey=r58lpoyt6xu5rtg291cv9fnei')
      fp_regression = h5py.File("test/model_regression_data.h5",'r') 
    regression_hash = md5("test/model_regression_data.h5")
    print("hash of model_regression_data.h5 is ",regression_hash)

  # remove models if you don't have them
  dont_test = [#"NRHybSur2dq15",
               #"BHPTNRSur1dq1e4",
               #"EMRISur1dq1e4",
               #"NRHybSur3dq8_CCE",
               "NRSur4d2s_TDROM_grid12", # 10 GB file
               "NRSur4d2s_FDROM_grid12", # 10 GB file
               #"SpEC_q1_10_NoSpin_linear_alt",
               #"SpEC_q1_10_NoSpin_linear",
               "EOBNRv2", #TODO: this is two surrogates in one. Break up?
               #"SpEC_q1_10_NoSpin",
               #"EOBNRv2_tutorial",
               #"NRHybSur3dq8",
               #"NRHybSur3dq8Tidal",
               #"NRSur7dq4"
               ]


  # Common directory where all surrogates are assumed to be located
  surrogate_path = gws.catalog.download_path()
 
  # repeatability can be useful for regression tests
  np.random.seed(0)

  # for each model, associate its surrogate data file
  models = [model for model in gws.catalog._surrogate_world]
  print(models)

  models_to_test = {}
  for model in models:
    surrogate_data = surrogate_path+os.path.basename(gws.catalog._surrogate_world[model][0])
    if os.path.isfile(surrogate_data): # surrogate data file exists
      models_to_test[model] = surrogate_data
    else: # file missing 
      msg = "WARNING: Surrogate missing!!!\n"
      msg += "Surrogate data assumed to be in the path %s.\n"%surrogate_data
      msg += "If the data is somewhere else, change the path or move the file.\n\n"
      msg +="To download this surrogate, from ipython do\n\n >>> gws.catalog.pull(%s)\n"%model
      print(msg)
      time.sleep(1)
      # Comment out following assert if experimenting with model-specific tests
      #assert(False)
      
 
  # also test the tutorial surrogate
  models_to_test["EOBNRv2_tutorial"] = gws.__path__[0] + "/../tutorial/TutorialSurrogate/EOB_q1_2_NoSpin_Mode22/"

  # remove models from testing...
  for i in dont_test:
    try:
      models_to_test.pop(i)
      print("model %s removed from testing"%i)
    except KeyError:
      print("model %s cannot be removed"%i)

  fp = h5py.File(h5_file,"w")

  # for each model, select three random points to evaluate at
  param_samples_tested = []
  for model, datafile in models_to_test.items():

    print("Generating data for model = %s"%model)
    print(datafile)

    if model in surrogate_old_interface:
      sur = gws.EvaluateSurrogate(datafile)
      # this is the surrogate's parameterization (e.g. log(q), \eta) region!
      # Technically should be using q. For SpEC_q1_10_NoSpin, Scott manually 
      # mapped eta to q, directly modifing the regression data file (10/5/2022)
      p_mins = sur.param_space.min_vals()
      p_maxs = sur.param_space.max_vals()
    elif model in surrogate_loader_interface:
      # sur = gws.LoadSurrogate(datafile) # tidal and aligned models use the same h5 file (so wrong one loaded) 
      sur = gws.LoadSurrogate(model)
      try:
        p_mins = sur._sur_dimless.param_space.min_vals()
        p_maxs = sur._sur_dimless.param_space.max_vals()
      except AttributeError: # NRSur7dq4 does not have object sur._sur_dimless.param_space
        p_mins = None
        p_maxs = None
    else:
      sur = surrogate.FastTensorSplineSurrogate()
      sur.load(datafile)
      p_mins = sur.param_space.min_vals()
      p_maxs = sur.param_space.max_vals()

    # NOTE: NRHybSur3dq8_CCE and NRHybSur3dq8 will report different max/min
    #       because NRHybSur3dq8 defines its interval in q while NRHybSur3dq8_CCE
    #       defines its interval in log(q). This change in the hdf5 file's
    #       behavior is due a change in convert_to_gwsurrogate.py (building code).
    #       Neither NRHybSur3dq8_CCE nor NRHybSur3dq8 use 
    #       sur._sur_dimless.param_space for model evaluations. See ParamSpace's docstring 
        
    print("parameter minimum values for model %s"%model,p_mins)
    print("parameter maximum values for model %s"%model,p_maxs)

    param_samples = []
    if generate_regression_data: # pick new points to compute regression data at
      for i in range(3):  # sample parameter space 3 times 
        if model in list(model_sampler.keys()):
          custom_sampler = model_sampler[model]
          x, tidOpts, pecOpts = custom_sampler(i) # [q, chiA, chiB], tidOpts
        else: # default sampler for spin-aligned BBH models
          x = [] # [q, chiAz, chiBz]
          for j in range(len(p_mins)):
            xj_min = p_mins[j]
            xj_max = p_maxs[j]
            tmp = float(np.random.uniform(xj_min, xj_max,size=1))
            x.append(tmp)
          tidOpts = None
          pecOpts = None
        param_samples.append([x, tidOpts, pecOpts])
    else: # get point at which to compute comparison waveform data
      for i in range(3):
        if model in list(model_sampler.keys()): # use sample points if provided 
          custom_sampler = model_sampler[model]
          x, tidOpts, pecOpts = custom_sampler(i) # [q, chiA, chiB], tidOpts
        else: # pull regression points from regression data file; none-tidal models only
          print(model+"/parameter%i/parameter"%i)
          x = []
          x.append( list(fp_regression[model+"/parameter%i/parameter"%i][:]) ) # [q, chiAz, chiBz]
          x = list(fp_regression[model+"/parameter%i/parameter"%i][:]) # [q, chiAz, chiBz]
          tidOpts = None
          pecOpts = None
        param_samples.append([x, tidOpts, pecOpts])

    param_samples_tested.append(param_samples)

    model_grp = fp.create_group(model)
    for i, ps in enumerate(param_samples):
      x = ps[0]
      tidOpts = ps[1]
      pecOpts = ps[2]
      if model in surrogate_old_interface:
        ps_float = x[0] # TODO: generalize interface 
        modes, t, hp, hc = sur(q=ps_float,mode_sum=False,fake_neg_modes=True)
      else:
        if model in surrogate_loader_interface:
          q = x[0]
          if type(x[1]) is np.float64 or type(x[1]) is float: # chiz
            chiA = np.array([0, 0, x[1]])
            chiB = np.array([0, 0, x[2]])
          elif len(x[1])==3: # spin vector
            chiA = np.array(x[1])
            chiB = np.array(x[2])
          else:
            raise ValueError
          try:
                  # Regression samples outside of the training interval.
                  # Warnings are raised (as they should) but could 
                  # appear bad to a new user
              if model in ["NRSur7dq4"]:
                with warnings.catch_warnings():
                  warnings.simplefilter("ignore")
                  t, h, dyanmics = sur(q, chiA, chiB, f_low=0.0, tidal_opts=tidOpts, precessing_opts=pecOpts)
              else:
                t, h, dyanmics = sur(q, chiA, chiB, f_low=0.0, tidal_opts=tidOpts, precessing_opts=pecOpts)
          except ValueError: # some models do not allow for f_low=0.0 and require a time step
             # step size, Units of M
             # initial frequency, Units of cycles/M
            t, h, dyanmics = sur(q, chiA, chiB, dt = 0.25, f_low=3.e-3, tidal_opts=tidOpts, precessing_opts=pecOpts) 
        else:
          h= sur(x)
        try:
          modes = sur.mode_list
          h_np = [h[mode] for mode in modes]
        except AttributeError: # for new interface
          modes = sur._sur_dimless.mode_list
          h_np = [h[mode] for mode in modes]

        h_np = np.vstack(h_np)
        hp = np.real(h_np)
        hc = np.imag(h_np)
      samplei = model_grp.create_group("parameter"+str(i))
      if model in list(model_sampler.keys()):
        # model samplers return a list of lists, which we flatten into 
        # a list of numbers for storing in an h5 dataset
        x = flatten_params(x)
      samplei.create_dataset("parameter",data=x)
      samplei.create_dataset("hp", data=hp, dtype='float32')
      samplei.create_dataset("hc", data=hc, dtype='float32')
      # added on 7/24/2022 -- not part of model_regression_data.h5
      # but is used in code logic below
      samplei.create_dataset("time", data=t, dtype='float32')
      samplei.create_dataset("modes", data=np.array(modes), dtype='int')

  fp.close()

  if not generate_regression_data:
    fp = h5py.File(h5_file,"r") # reopen comparison data
    for model in models_to_test.keys():
      print("testing model %s ..."%model)
      for i in range(3): # 3 parameter samples
        hp_regression = fp_regression[model+"/parameter%i/hp"%i][:]
        hc_regression = fp_regression[model+"/parameter%i/hc"%i][:]
        hp_comparison = fp[model+"/parameter%i/hp"%i][:]
        hc_comparison = fp[model+"/parameter%i/hc"%i][:]

        # only test data up to 90M past merger
        # Currently not used (but keep in case this is reactivated)
        # t_indx = np.argmin(np.abs(fp[model+"/parameter%i/time"%i][:] - 90) )
        t_indx = fp[model+"/parameter%i/time"%i][:].shape[0]  # use all time points

        # model-specific relative tolerances. This is needed because certain models
        # have dependencies (e.g. GSL or sklearn) that will break our tests!
        local_rtol = rtols.get(model, rtol_default)

        print("Model %s uses a relative error tolerance of %e"%(model,local_rtol))

        for j, mode in enumerate(fp[model+"/parameter%i/modes"%i][:]): # test mode-by-mode
          err_msg="Failed: model %s for mode index %i (ell = %i,m = %i)"%(model,j,mode[0],mode[1])

          # test hp and hc separately 
          # Note: because hp and hc oscillate about 0, this test can easily fail (relative error check) 
          # if hp/hc change by very small amounts
          #np.testing.assert_allclose(hp_regression[j,:t_indx], hp_comparison[j,:t_indx], rtol=local_rtol, atol=atol, err_msg=err_msg)
          #np.testing.assert_allclose(hc_regression[j,:t_indx], hc_comparison[j,:t_indx], rtol=local_rtol, atol=atol, err_msg=err_msg)

          # test complexified h.
          # This is a more robust relative-error test because the |h| does not pass through 0
          h_regression = hp_regression[j,:t_indx] + 1.0j*hc_regression[j,:t_indx]
          h_comparison = hp_comparison[j,:t_indx] + 1.0j*hc_comparison[j,:t_indx]
          np.testing.assert_allclose(h_regression, h_comparison, rtol=local_rtol, atol=atol, err_msg=err_msg)


        # when debugging the tests, it can be useful to dump the data
        #np.save("hp_regression-%i.npy"%i,hp_regression)
        #np.save("hp_comparison-%i.npy"%i,hp_comparison)
        #np.save("hc_regression-%i.npy"%i,hc_regression)
        #np.save("hc_comparison-%i.npy"%i,hc_comparison)


    # fails due to round-off error differences of different machines
    #fp_regression.close()
    #process = subprocess.Popen(["h5diff", "test/model_regression_data.h5",h5_file],
    #                           stdin=subprocess.PIPE,
    #                           stdout=subprocess.PIPE,
    #                           stderr=subprocess.PIPE)
    #returncode = process.wait()
    #stdout, stderr = process.communicate()
    #if returncode == 0:
    #    assert(True)
    #else:
    #    print(stdout)
    #    print(stderr)
    #    assert(False)
  print("models tested... ")
  #for i, model_tested in enumerate(models_to_test.keys()):
  #  print("model %s at points..."%model_tested+str(param_samples_tested[i]))


#------------------------------------------------------------------------------
if __name__ == "__main__":
  test_model_regression(generate_regression_data=True)



"""
Tensor spline evaluation code.
The intended use case is when we have very many different splines which
all have the same domain (breakpoints). Careful usage of numpy keeps the
added computational cost of using python vs. (for example) C with gsl
minimal when evaluating all splines simultaneously at the same parameter
value.
"""

from __future__ import division  # for py2

__copyright__ = "Copyright (C) 2014 Scott Field and Chad Galley"
__email__ = "sfield@astro.cornell.edu, crgalley@tapir.caltech.edu"
__status__ = "testing"
__author__ = "Jonathan Blackman"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

# print("__name__ = %s"%__name__)
# print("__package__= %s"%__package__)

if (
    "__package__" not in locals() or __package__ is None or __package__ == ""
):  # py2 and py3 compatible
    print("setting __package__ to gwsurrogate.new so relative imports work")
    __package__ = "gwsurrogate.new"

print("__name__ = %s" % __name__)
print("__package__= %s" % __package__)

import numpy as np
import h5py
from .saveH5Object import SimpleH5Object
import itertools
from functools import reduce


def _cubic_spline_breaks(knot_vec):
    """
    Given spline knots [x0, x1, ..., xN-1=xf], we define the breakpoints as
    [x0, x0, x0, x0, x1, x2, ..., xN-2, xf, xf, xf, xf]
    which has length N+6.
    Note that there are N+2 spline coefficients corresponding to bsplines where
    coef[i] corresponds to the bspline with support between breakpoints[i] and
    breakpoints[i+4], so the supports are:
        i=0:    [x0, x1)
        i=1:    [x0, x2)
        i=2:    [x0, x3)
        i=3:    [x0, x4)
        i=4:    [x1, x5)
        ...
        i=N-3:  [xN-6, xN-2)
        i=N-2:  [xN-5, xf)
        i=N-1:  [xN-4, xf)
        i=N:    [xN-3, xf)
        i=N+1:  [xN-2, xf)
    """

    if np.min(np.diff(knot_vec)) < 0.0:
        raise Exception("knot_vec should be strictly increasing")

    return np.append(
        np.append(np.ones(3) * knot_vec[0], knot_vec), np.ones(3) * knot_vec[-1]
    )


def cubic_spline_breaksToknots(bvec):
    """
    Given breakpoints generated from _cubic_spline_breaks,
    [x0, x0, x0, x0, x1, x2, ..., xN-2, xf, xf, xf, xf],
    return the spline knots [x0, x1, ..., xN-1=xf].
    This function ``undoes" _cubic_spline_breaks:
    knot_vec = _cubic_spline_breaks2knots(_cubic_spline_breaks(knot_vec))
    """

    return bvec[3:-3]


# -----------------------------------------------------------------------------


def _cubic_bspline_eval_nonzero_1d(x, bvec, tol=1.0e-12):
    """
    Given a point x and an array of breakpoints bvec, determine the 4
    potentially non-zero bsplines and evaluate them at x.
    Returns i0, bspline_evals.
    """

    knots = bvec[3:-3]

    # Nudge parameters if barely outside domain
    if knots[0] - tol < x <= knots[0]:
        x = knots[0]
    if knots[-1] - tol < x < knots[-1] + tol:
        # The domain is [knots[0], knots[-1]) so nudge inwards a little
        x = knots[-1] - tol

    if x < knots[0] or x > knots[-1]:
        raise Exception(
            "%s outside of [%s, %s] parameter range!" % (x, knots[0], knots[-1])
        )

    if x <= knots[0]:
        i0 = 0
    elif x >= knots[-1]:
        i0 = len(knots) - 2
    else:
        i0 = np.where(x >= knots)[0][-1]

    return i0, np.array(
        [_bspline_eval(x, bvec[i : i + 5], k=3) for i in range(i0, i0 + 4)]
    )


# -----------------------------------------------------------------------------


def _bspline_eval(x, bvec, k, check=True):
    if check and np.min(np.diff(bvec)) < 0.0:
        raise Exception("bvec should be strictly increasing")

    if x < bvec[0] or x >= bvec[-1]:
        return 0.0

    if k == 0:
        return 1.0

    c1 = _bspline_eval(x, bvec[:-1], k - 1, check=False)
    c2 = _bspline_eval(x, bvec[1:], k - 1, check=False)
    if abs(c1) > 0.0:
        c1 *= (x - bvec[0]) / (bvec[-2] - bvec[0])
    if abs(c2) > 0.0:
        c2 *= (bvec[-1] - x) / (bvec[-1] - bvec[1])
    return c1 + c2


def memoize_spline_call(func):
    """Decorator for TensorSplineGrid's __call__ method.

    Multiple calls to __call__ typically have the same xvec
    value, in which case we can return the last result without
    recomputing anything.

    A decorator is needed because member variables of TensorSplineGrid
    are expected to be loaded from an hdf5 file."""

    last_call = [np.nan]
    last_return = [np.nan, np.nan, np.nan]

    def decorated_function(self, xvec):
        # TODO: make this next code block its own function? Used elsewhere
        xshape = np.shape(xvec)

        # It's convenient to be able to accept a float instead of a length-1
        # array for 1d parameter spaces.

        if len(xshape) == 0:
            xvec = np.array([xvec])
            xshape = np.shape(xvec)

        if np.max(np.abs(last_call[0] - xvec)) != 0:
            last_call[0] = xvec
            last_return[0], last_return[1], last_return[2] = func(self, xvec)
        return last_return[0], last_return[1], last_return[2]

    return decorated_function


class TensorSplineGrid(SimpleH5Object):
    def __init__(self, knot_vecs=[]):
        """
        Stores the geometric information needed for tensor-spline interpolation.
        No coefficients are stored, so one TensorSplineGrid can be used for many
        splines sharing a common grid.
        """

        super(TensorSplineGrid, self).__init__()

        self.dim = len(knot_vecs)
        self.grid_dim = [len(v) for v in knot_vecs]
        self.breakpoint_vecs = [_cubic_spline_breaks(v) for v in knot_vecs]

    def bspline_eval_nonzero(self, xvec):
        """
        Returns imin_vals, spline_evals.
        Each have length self.dim.
        imin_vals gives the minimum spline coefficient index for the spline_evals.
        Each element of spline_evals consists of the 4 potentially non-zero
            bspline evaluations.
        """

        res = [
            _cubic_bspline_eval_nonzero_1d(x, bvec)
            for x, bvec in zip(xvec, self.breakpoint_vecs)
        ]

        imin_vals, spline_evals = [list(t) for t in zip(*res)]
        return imin_vals, spline_evals

    @memoize_spline_call
    def __call__(self, xvec):
        """
        Evaluates potentially non-zero spline basis function products.
        xvec: The point in parameter space. Must be of type numpy.ndarray
        Returns:
            eval_prods: Products of spline evaluations in all parameter space
                        directions, which can be summed up with spline coefficients.
            sl: Single slice so we can make better use of np.sum:
            summed_axes: See below

            With these three items, we can evaluate a spline from its coefficients:

               >>> y = np.sum(c[sl] * eval_prods, axis=summed_axes)

            See FastTensorSplineSurrogate's call method

        """

        # All splines use the same grid, so we determine the 4^d potentially
        # non-zero spline basis function products here which can be used for
        # all interpolations.
        imin_vals, spline_evals = self.bspline_eval_nonzero(xvec)

        # Create 4^d hypercube of bspline products
        # Reduce uses left to right, transposes sound ugly, just reverse.
        # TODO: It could be faster to use np.ravel(np.outer(...))
        eval_prods = reduce(
            lambda a, b: np.array([a * x for x in b]), spline_evals[::-1]
        )

        # This slice can be passed to a numpy grid of spline coefficients
        # to extract the 4^d relevant coefficients
        sl_base = tuple(slice(i0, i0 + 4, None) for i0 in imin_vals)

        # We will have arrays of shape (n_EI, n1, n2, ..., nd) where n_EI
        # is the number of empirical nodes, and n1, ..., nd are the number of
        # spline coefficients (2 + the number of grid points) in each dimension.
        # We want to sum over everything except the first index to evaluate the
        # n_EI different splines.
        summed_axes = tuple(i + 1 for i in range(self.dim))

        # Build a single slice so we can make better use of np.sum:
        # NOTE: sl object is meant to be used with spline coefficients
        # for multiple EIM nodes
        sl = tuple(itertools.chain([slice(None)], sl_base))

        # self.last_return = [eval_prods, sl, summed_axes]

        return eval_prods, sl, summed_axes


# -----------------------------------------------------------------------------


# TODO: maybe ts_grid(x) should return sl_base and build sl here.
# the sl oject should also get cached
def fast_tensor_spline_eval(x, ts_grid, spline_coeffs):
    """Evaluate SPLINE_COEFFS defined on the grid TS_GRID
    at an n-dimensional value X."""

    # See TensorSplineGrid's call method for documentation
    eval_prods, sl, summed_axes = ts_grid(x)

    return np.sum(spline_coeffs[sl] * eval_prods, axis=summed_axes)


def fast_complex_tensor_spline_eval(x, ts_grid, spline_coeffs_real, spline_coeffs_imag):
    """Evaluate SPLINE_COEFFS_REAL and SPLINE_COEFFS_IMAG defined on the
    grid TS_GRID at an n-dimensional value X."""

    nre = fast_tensor_spline_eval(x, ts_grid, spline_coeffs_real)
    nim = fast_tensor_spline_eval(x, ts_grid, spline_coeffs_imag)

    return nre + 1.0j * nim
